@@ -1,4 +1,4 @@
-/*	$NetBSD: read.c,v 1.107 2021/08/15 10:08:41 christos Exp $	*/
+/*	$NetBSD: read.c,v 1.39 2005/08/02 12:11:14 christos Exp $	*/
 
 /*-
  * Copyright (c) 1992, 1993
@@ -37,90 +37,47 @@
 #if 0
 static char sccsid[] = "@(#)read.c	8.1 (Berkeley) 6/4/93";
 #else
-__RCSID("$NetBSD: read.c,v 1.107 2021/08/15 10:08:41 christos Exp $");
+__RCSID("$NetBSD: read.c,v 1.39 2005/08/02 12:11:14 christos Exp $");
 #endif
 #endif /* not lint && not SCCSID */
 
 /*
- * read.c: Terminal read functions
+ * read.c: Clean this junk up! This is horrible code.
+ *	   Terminal read functions
  */
-#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <limits.h>
-#include <stdlib.h>
-#include <string.h>
 #include <unistd.h>
-
+#include <stdlib.h>
 #include "el.h"
-#include "fcns.h"
-#include "read.h"
 
-#define	EL_MAXMACRO	10
+#define	OKCMD	-1
 
-struct macros {
-	wchar_t	**macro;
-	int	  level;
-	int	  offset;
-};
-
-struct el_read_t {
-	struct macros	 macros;
-	el_rfunc_t	 read_char;	/* Function to read a character. */
-	int		 read_errno;
-};
-
-static int	read__fixio(int, int);
-static int	read_char(EditLine *, wchar_t *);
-static int	read_getcmd(EditLine *, el_action_t *, wchar_t *);
-static void	read_clearmacros(struct macros *);
-static void	read_pop(struct macros *);
-static const wchar_t *noedit_wgets(EditLine *, int *);
+private int	read__fixio(int, int);
+private int	read_preread(EditLine *);
+private int	read_char(EditLine *, char *);
+private int	read_getcmd(EditLine *, el_action_t *, char *);
 
 /* read_init():
  *	Initialize the read stuff
  */
-libedit_private int
+protected int
 read_init(EditLine *el)
 {
-	struct macros *ma;
-
-	if ((el->el_read = el_malloc(sizeof(*el->el_read))) == NULL)
-		return -1;
-
-	ma = &el->el_read->macros;
-	if ((ma->macro = el_calloc(EL_MAXMACRO, sizeof(*ma->macro))) == NULL) {
-		free(el->el_read);
-		return -1;
-	}
-	ma->level = -1;
-	ma->offset = 0;
-
 	/* builtin read_char */
-	el->el_read->read_char = read_char;
+	el->el_read.read_char = read_char;
 	return 0;
 }
 
-/* el_read_end():
- *	Free the data structures used by the read stuff.
- */
-libedit_private void
-read_end(struct el_read_t *el_read)
-{
-	read_clearmacros(&el_read->macros);
-	el_free(el_read->macros.macro);
-	el_read->macros.macro = NULL;
-	el_free(el_read);
-}
 
 /* el_read_setfn():
  *	Set the read char function to the one provided.
  *	If it is set to EL_BUILTIN_GETCFN, then reset to the builtin one.
  */
-libedit_private int
-el_read_setfn(struct el_read_t *el_read, el_rfunc_t rc)
+protected int
+el_read_setfn(EditLine *el, el_rfunc_t rc)
 {
-	el_read->read_char = (rc == EL_BUILTIN_GETCFN) ? read_char : rc;
+	el->el_read.read_char = (rc == EL_BUILTIN_GETCFN) ? read_char : rc;
 	return 0;
 }
 
@@ -129,19 +86,42 @@ el_read_setfn(struct el_read_t *el_read, el_rfunc_t rc)
  *	return the current read char function, or EL_BUILTIN_GETCFN
  *	if it is the default one
  */
-libedit_private el_rfunc_t
-el_read_getfn(struct el_read_t *el_read)
+protected el_rfunc_t
+el_read_getfn(EditLine *el)
 {
-       return el_read->read_char == read_char ?
-	    EL_BUILTIN_GETCFN : el_read->read_char;
+       return (el->el_read.read_char == read_char) ?
+	    EL_BUILTIN_GETCFN : el->el_read.read_char;
 }
 
 
+#ifndef MIN
+#define MIN(A,B) ((A) < (B) ? (A) : (B))
+#endif
+
+#ifdef DEBUG_EDIT
+private void
+read_debug(EditLine *el)
+{
+
+	if (el->el_line.cursor > el->el_line.lastchar)
+		(void) fprintf(el->el_errfile, "cursor > lastchar\r\n");
+	if (el->el_line.cursor < el->el_line.buffer)
+		(void) fprintf(el->el_errfile, "cursor < buffer\r\n");
+	if (el->el_line.cursor > el->el_line.limit)
+		(void) fprintf(el->el_errfile, "cursor > limit\r\n");
+	if (el->el_line.lastchar > el->el_line.limit)
+		(void) fprintf(el->el_errfile, "lastchar > limit\r\n");
+	if (el->el_line.limit != &el->el_line.buffer[EL_BUFSIZ - 2])
+		(void) fprintf(el->el_errfile, "limit != &buffer[EL_BUFSIZ-2]\r\n");
+}
+#endif /* DEBUG_EDIT */
+
+
 /* read__fixio():
  *	Try to recover from a read error
  */
 /* ARGSUSED */
-static int
+private int
 read__fixio(int fd __attribute__((__unused__)), int e)
 {
 
@@ -151,7 +131,7 @@ read__fixio(int fd __attribute__((__unused__)), int e)
 #ifdef EWOULDBLOCK
 	case EWOULDBLOCK:
 #ifndef TRY_AGAIN
-#define TRY_AGAIN
+#define	TRY_AGAIN
 #endif
 #endif /* EWOULDBLOCK */
 
@@ -159,7 +139,7 @@ read__fixio(int fd __attribute__((__unused__)), int e)
 #if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
 	case EAGAIN:
 #ifndef TRY_AGAIN
-#define TRY_AGAIN
+#define	TRY_AGAIN
 #endif
 #endif /* EWOULDBLOCK && EWOULDBLOCK != EAGAIN */
 #endif /* POSIX && EAGAIN */
@@ -168,10 +148,10 @@ read__fixio(int fd __attribute__((__unused__)), int e)
 #ifdef TRY_AGAIN
 #if defined(F_SETFL) && defined(O_NDELAY)
 		if ((e = fcntl(fd, F_GETFL, 0)) == -1)
-			return -1;
+			return (-1);
 
 		if (fcntl(fd, F_SETFL, e & ~O_NDELAY) == -1)
-			return -1;
+			return (-1);
 		else
 			e = 1;
 #endif /* F_SETFL && O_NDELAY */
@@ -180,61 +160,88 @@ read__fixio(int fd __attribute__((__unused__)), int e)
 		{
 			int zero = 0;
 
-			if (ioctl(fd, FIONBIO, &zero) == -1)
-				return -1;
+			if (ioctl(fd, FIONBIO, (ioctl_t) & zero) == -1)
+				return (-1);
 			else
 				e = 1;
 		}
 #endif /* FIONBIO */
 
 #endif /* TRY_AGAIN */
-		return e ? 0 : -1;
+		return (e ? 0 : -1);
 
 	case EINTR:
-		return 0;
+		return (0);
 
 	default:
-		return -1;
+		return (-1);
 	}
 }
 
 
+/* read_preread():
+ *	Try to read the stuff in the input queue;
+ */
+private int
+read_preread(EditLine *el)
+{
+	int chrs = 0;
+
+	if (el->el_tty.t_mode == ED_IO)
+		return (0);
+
+#ifdef FIONREAD
+	(void) ioctl(el->el_infd, FIONREAD, (ioctl_t) & chrs);
+	if (chrs > 0) {
+		char buf[EL_BUFSIZ];
+
+		chrs = read(el->el_infd, buf,
+		    (size_t) MIN(chrs, EL_BUFSIZ - 1));
+		if (chrs > 0) {
+			buf[chrs] = '\0';
+			el_push(el, buf);
+		}
+	}
+#endif /* FIONREAD */
+
+	return (chrs > 0);
+}
+
+
 /* el_push():
  *	Push a macro
  */
-void
-el_wpush(EditLine *el, const wchar_t *str)
+public void
+el_push(EditLine *el, char *str)
 {
-	struct macros *ma = &el->el_read->macros;
+	c_macro_t *ma = &el->el_chared.c_macro;
 
 	if (str != NULL && ma->level + 1 < EL_MAXMACRO) {
 		ma->level++;
-		if ((ma->macro[ma->level] = wcsdup(str)) != NULL)
+		if ((ma->macro[ma->level] = el_strdup(str)) != NULL)
 			return;
 		ma->level--;
 	}
-	terminal_beep(el);
-	terminal__flush(el);
+	term_beep(el);
+	term__flush();
 }
 
 
 /* read_getcmd():
- *	Get next command from the input stream,
- *	return 0 on success or -1 on EOF or error.
- *	Character values > 255 are not looked up in the map, but inserted.
+ *	Return next command from the input stream.
  */
-static int
-read_getcmd(EditLine *el, el_action_t *cmdnum, wchar_t *ch)
+private int
+read_getcmd(EditLine *el, el_action_t *cmdnum, char *ch)
 {
-	static const wchar_t meta = (wchar_t)0x80;
 	el_action_t cmd;
+	int num;
 
 	do {
-		if (el_wgetc(el, ch) != 1)
-			return -1;
+		if ((num = el_getc(el, ch)) != 1)	/* if EOF or error */
+			return (num);
 
 #ifdef	KANJI
-		if ((*ch & meta)) {
+		if ((*ch & 0200)) {
 			el->el_state.metanext = 0;
 			cmd = CcViMap[' '];
 			break;
@@ -243,180 +250,107 @@ read_getcmd(EditLine *el, el_action_t *cmdnum, wchar_t *ch)
 
 		if (el->el_state.metanext) {
 			el->el_state.metanext = 0;
-			*ch |= meta;
+			*ch |= 0200;
 		}
-		if (*ch >= N_KEYS)
-			cmd = ED_INSERT;
-		else
-			cmd = el->el_map.current[(unsigned char) *ch];
+		cmd = el->el_map.current[(unsigned char) *ch];
 		if (cmd == ED_SEQUENCE_LEAD_IN) {
-			keymacro_value_t val;
-			switch (keymacro_get(el, ch, &val)) {
+			key_value_t val;
+			switch (key_get(el, ch, &val)) {
 			case XK_CMD:
 				cmd = val.cmd;
 				break;
 			case XK_STR:
-				el_wpush(el, val.str);
+				el_push(el, val.str);
 				break;
-			case XK_NOD:
-				return -1;
+#ifdef notyet
+			case XK_EXE:
+				/* XXX: In the future to run a user function */
+				RunCommand(val.str);
+				break;
+#endif
 			default:
 				EL_ABORT((el->el_errfile, "Bad XK_ type \n"));
 				break;
 			}
 		}
+		if (el->el_map.alt == NULL)
+			el->el_map.current = el->el_map.key;
 	} while (cmd == ED_SEQUENCE_LEAD_IN);
 	*cmdnum = cmd;
-	return 0;
+	return (OKCMD);
 }
 
+
 /* read_char():
  *	Read a character from the tty.
  */
-static int
-read_char(EditLine *el, wchar_t *cp)
+private int
+read_char(EditLine *el, char *cp)
 {
-	ssize_t num_read;
-	int tried = (el->el_flags & FIXIO) == 0;
-	char cbuf[MB_LEN_MAX];
-	size_t cbp = 0;
-	int save_errno = errno;
-
- again:
-	el->el_signal->sig_no = 0;
-	while ((num_read = read(el->el_infd, cbuf + cbp, (size_t)1)) == -1) {
-		int e = errno;
-		switch (el->el_signal->sig_no) {
-		case SIGCONT:
-			el_wset(el, EL_REFRESH);
-			/*FALLTHROUGH*/
-		case SIGWINCH:
-			sig_set(el);
-			goto again;
-		default:
-			break;
-		}
-		if (!tried && read__fixio(el->el_infd, e) == 0) {
-			errno = save_errno;
-			tried = 1;
-		} else {
-			errno = e;
-			*cp = L'\0';
-			return -1;
-		}
-	}
-
-	/* Test for EOF */
-	if (num_read == 0) {
-		*cp = L'\0';
-		return 0;
-	}
-
-	for (;;) {
-		mbstate_t mbs;
-
-		++cbp;
-		/* This only works because UTF8 is stateless. */
-		memset(&mbs, 0, sizeof(mbs));
-		switch (mbrtowc(cp, cbuf, cbp, &mbs)) {
-		case (size_t)-1:
-			if (cbp > 1) {
-				/*
-				 * Invalid sequence, discard all bytes
-				 * except the last one.
-				 */
-				cbuf[0] = cbuf[cbp - 1];
-				cbp = 0;
-				break;
-			} else {
-				/* Invalid byte, discard it. */
-				cbp = 0;
-				goto again;
-			}
-		case (size_t)-2:
-			if (cbp >= MB_LEN_MAX) {
-				errno = EILSEQ;
-				*cp = L'\0';
-				return -1;
-			}
-			/* Incomplete sequence, read another byte. */
-			goto again;
-		default:
-			/* Valid character, process it. */
-			return 1;
-		}
-	}
-}
-
-/* read_pop():
- *	Pop a macro from the stack
- */
-static void
-read_pop(struct macros *ma)
-{
-	int i;
-
-	el_free(ma->macro[0]);
-	for (i = 0; i < ma->level; i++)
-		ma->macro[i] = ma->macro[i + 1];
-	ma->level--;
-	ma->offset = 0;
-}
-
-static void
-read_clearmacros(struct macros *ma)
-{
-	while (ma->level >= 0)
-		el_free(ma->macro[ma->level--]);
-	ma->offset = 0;
-}
-
-/* el_wgetc():
- *	Read a wide character
- */
-int
-el_wgetc(EditLine *el, wchar_t *cp)
-{
-	struct macros *ma = &el->el_read->macros;
 	int num_read;
+	int tried = 0;
 
-	terminal__flush(el);
+	while ((num_read = read(el->el_infd, cp, 1)) == -1)
+		if (!tried && read__fixio(el->el_infd, errno) == 0)
+			tried = 1;
+		else {
+			*cp = '\0';
+			return (-1);
+		}
+
+	return (num_read);
+}
+
+
+/* el_getc():
+ *	Read a character
+ */
+public int
+el_getc(EditLine *el, char *cp)
+{
+	int num_read;
+	c_macro_t *ma = &el->el_chared.c_macro;
+
+	term__flush();
 	for (;;) {
+		if (ma->level < 0) {
+			if (!read_preread(el))
+				break;
+		}
 		if (ma->level < 0)
 			break;
 
-		if (ma->macro[0][ma->offset] == '\0') {
-			read_pop(ma);
+		if (ma->macro[ma->level][ma->offset] == '\0') {
+			el_free(ma->macro[ma->level--]);
+			ma->offset = 0;
 			continue;
 		}
-
-		*cp = ma->macro[0][ma->offset++];
-
-		if (ma->macro[0][ma->offset] == '\0') {
+		*cp = ma->macro[ma->level][ma->offset++] & 0377;
+		if (ma->macro[ma->level][ma->offset] == '\0') {
 			/* Needed for QuoteMode On */
-			read_pop(ma);
+			el_free(ma->macro[ma->level--]);
+			ma->offset = 0;
 		}
-
-		return 1;
+		return (1);
 	}
 
+#ifdef DEBUG_READ
+	(void) fprintf(el->el_errfile, "Turning raw mode on\n");
+#endif /* DEBUG_READ */
 	if (tty_rawmode(el) < 0)/* make sure the tty is set up correctly */
-		return 0;
+		return (0);
 
-	num_read = (*el->el_read->read_char)(el, cp);
-
-	/*
-	 * Remember the original reason of a read failure
-	 * such that el_wgets() can restore it after doing
-	 * various cleanup operation that might change errno.
-	 */
-	if (num_read < 0)
-		el->el_read->read_errno = errno;
-
-	return num_read;
+#ifdef DEBUG_READ
+	(void) fprintf(el->el_errfile, "Reading a character\n");
+#endif /* DEBUG_READ */
+	num_read = (*el->el_read.read_char)(el, cp);
+#ifdef DEBUG_READ
+	(void) fprintf(el->el_errfile, "Got it %c\n", *cp);
+#endif /* DEBUG_READ */
+	return (num_read);
 }
 
-libedit_private void
+protected void
 read_prepare(EditLine *el)
 {
 	if (el->el_flags & HANDLE_SIGNALS)
@@ -430,14 +364,14 @@ read_prepare(EditLine *el)
 	   we have the wrong size. */
 	el_resize(el);
 	re_clear_display(el);	/* reset the display stuff */
-	ch_reset(el);
+	ch_reset(el, 0);
 	re_refresh(el);		/* print the prompt */
 
 	if (el->el_flags & UNBUFFERED)
-		terminal__flush(el);
+		term__flush();
 }
 
-libedit_private void
+protected void
 read_finish(EditLine *el)
 {
 	if ((el->el_flags & UNBUFFERED) == 0)
@@ -446,59 +380,55 @@ read_finish(EditLine *el)
 		sig_clr(el);
 }
 
-static const wchar_t *
-noedit_wgets(EditLine *el, int *nread)
-{
-	el_line_t	*lp = &el->el_line;
-	int		 num;
-
-	while ((num = (*el->el_read->read_char)(el, lp->lastchar)) == 1) {
-		if (lp->lastchar + 1 >= lp->limit &&
-		    !ch_enlargebufs(el, (size_t)2))
-			break;
-		lp->lastchar++;
-		if (el->el_flags & UNBUFFERED ||
-		    lp->lastchar[-1] == '\r' ||
-		    lp->lastchar[-1] == '\n')
-			break;
-	}
-	if (num == -1 && errno == EINTR)
-		lp->lastchar = lp->buffer;
-	lp->cursor = lp->lastchar;
-	*lp->lastchar = '\0';
-	*nread = (int)(lp->lastchar - lp->buffer);
-	return *nread ? lp->buffer : NULL;
-}
-
-const wchar_t *
-el_wgets(EditLine *el, int *nread)
+public const char *
+el_gets(EditLine *el, int *nread)
 {
 	int retval;
 	el_action_t cmdnum = 0;
 	int num;		/* how many chars we have read at NL */
-	wchar_t ch;
-	int nrb;
-
-	if (nread == NULL)
-		nread = &nrb;
-	*nread = 0;
-	el->el_read->read_errno = 0;
+	char ch;
+	int crlf = 0;
+#ifdef FIONREAD
+	c_macro_t *ma = &el->el_chared.c_macro;
+#endif /* FIONREAD */
 
 	if (el->el_flags & NO_TTY) {
-		el->el_line.lastchar = el->el_line.buffer;
-		return noedit_wgets(el, nread);
+		char *cp = el->el_line.buffer;
+		size_t idx;
+
+		while ((*el->el_read.read_char)(el, cp) == 1) {
+			/* make sure there is space for next character */
+			if (cp + 1 >= el->el_line.limit) {
+				idx = (cp - el->el_line.buffer);
+				if (!ch_enlargebufs(el, 2))
+					break;
+				cp = &el->el_line.buffer[idx];
+			}
+			cp++;
+			if (el->el_flags & UNBUFFERED)
+				break;
+			if (cp[-1] == '\r' || cp[-1] == '\n')
+				break;
+		}
+
+		el->el_line.cursor = el->el_line.lastchar = cp;
+		*cp = '\0';
+		if (nread)
+			*nread = el->el_line.cursor - el->el_line.buffer;
+		return (el->el_line.buffer);
 	}
 
-#ifdef FIONREAD
-	if (el->el_tty.t_mode == EX_IO && el->el_read->macros.level < 0) {
-		int chrs = 0;
 
-		(void) ioctl(el->el_infd, FIONREAD, &chrs);
+#ifdef FIONREAD
+	if (el->el_tty.t_mode == EX_IO && ma->level < 0) {
+		long chrs = 0;
+
+		(void) ioctl(el->el_infd, FIONREAD, (ioctl_t) & chrs);
 		if (chrs == 0) {
 			if (tty_rawmode(el) < 0) {
-				errno = 0;
-				*nread = 0;
-				return NULL;
+				if (nread)
+					*nread = 0;
+				return (NULL);
 			}
 		}
 	}
@@ -508,19 +438,75 @@ el_wgets(EditLine *el, int *nread)
 		read_prepare(el);
 
 	if (el->el_flags & EDIT_DISABLED) {
+		char *cp;
+		size_t idx;
 		if ((el->el_flags & UNBUFFERED) == 0)
-			el->el_line.lastchar = el->el_line.buffer;
-		terminal__flush(el);
-		return noedit_wgets(el, nread);
+			cp = el->el_line.buffer;
+		else
+			cp = el->el_line.lastchar;
+
+		term__flush();
+
+		while ((*el->el_read.read_char)(el, cp) == 1) {
+			/* make sure there is space next character */
+			if (cp + 1 >= el->el_line.limit) {
+				idx = (cp - el->el_line.buffer);
+				if (!ch_enlargebufs(el, 2))
+					break;
+				cp = &el->el_line.buffer[idx];
+			}
+			if (*cp == 4)	/* ought to be stty eof */
+				break;
+			cp++;
+			crlf = cp[-1] == '\r' || cp[-1] == '\n';
+			if (el->el_flags & UNBUFFERED)
+				break;
+			if (crlf)
+				break;
+		}
+
+		el->el_line.cursor = el->el_line.lastchar = cp;
+		*cp = '\0';
+		if (nread)
+			*nread = el->el_line.cursor - el->el_line.buffer;
+		return (el->el_line.buffer);
 	}
 
-	for (num = -1; num == -1;) {  /* while still editing this line */
+	for (num = OKCMD; num == OKCMD;) {	/* while still editing this
+						 * line */
+#ifdef DEBUG_EDIT
+		read_debug(el);
+#endif /* DEBUG_EDIT */
 		/* if EOF or error */
-		if (read_getcmd(el, &cmdnum, &ch) == -1)
+		if ((num = read_getcmd(el, &cmdnum, &ch)) != OKCMD) {
+#ifdef DEBUG_READ
+			(void) fprintf(el->el_errfile,
+			    "Returning from el_gets %d\n", num);
+#endif /* DEBUG_READ */
 			break;
-		if ((size_t)cmdnum >= el->el_map.nfunc) /* BUG CHECK command */
+		}
+		if ((unsigned int)cmdnum >= el->el_map.nfunc) {	/* BUG CHECK command */
+#ifdef DEBUG_EDIT
+			(void) fprintf(el->el_errfile,
+			    "ERROR: illegal command from key 0%o\r\n", ch);
+#endif /* DEBUG_EDIT */
 			continue;	/* try again */
+		}
 		/* now do the real command */
+#ifdef DEBUG_READ
+		{
+			el_bindings_t *b;
+			for (b = el->el_map.help; b->name; b++)
+				if (b->func == cmdnum)
+					break;
+			if (b->name)
+				(void) fprintf(el->el_errfile,
+				    "Executing %s\n", b->name);
+			else
+				(void) fprintf(el->el_errfile,
+				    "Error command = %d\n", cmdnum);
+		}
+#endif /* DEBUG_READ */
 		/* vi redo needs these way down the levels... */
 		el->el_state.thiscmd = cmdnum;
 		el->el_state.thisch = ch;
@@ -529,12 +515,16 @@ el_wgets(EditLine *el, int *nread)
 		    el->el_chared.c_redo.pos < el->el_chared.c_redo.lim) {
 			if (cmdnum == VI_DELETE_PREV_CHAR &&
 			    el->el_chared.c_redo.pos != el->el_chared.c_redo.buf
-			    && iswprint(el->el_chared.c_redo.pos[-1]))
+			    && isprint((unsigned char)el->el_chared.c_redo.pos[-1]))
 				el->el_chared.c_redo.pos--;
 			else
 				*el->el_chared.c_redo.pos++ = ch;
 		}
 		retval = (*el->el_map.func[cmdnum]) (el, ch);
+#ifdef DEBUG_READ
+		(void) fprintf(el->el_errfile,
+			"Returned state %d\n", retval );
+#endif /* DEBUG_READ */
 
 		/* save the last command here */
 		el->el_state.lastcmd = cmdnum;
@@ -556,7 +546,7 @@ el_wgets(EditLine *el, int *nread)
 
 		case CC_REFRESH_BEEP:
 			re_refresh(el);
-			terminal_beep(el);
+			term_beep(el);
 			break;
 
 		case CC_NORM:	/* normal char */
@@ -577,21 +567,28 @@ el_wgets(EditLine *el, int *nread)
 			break;
 
 		case CC_NEWLINE:	/* normal end of line */
-			num = (int)(el->el_line.lastchar - el->el_line.buffer);
+			num = el->el_line.lastchar - el->el_line.buffer;
 			break;
 
 		case CC_FATAL:	/* fatal error, reset to known state */
+#ifdef DEBUG_READ
+			(void) fprintf(el->el_errfile,
+			    "*** editor fatal ERROR ***\r\n\n");
+#endif /* DEBUG_READ */
 			/* put (real) cursor in a known place */
 			re_clear_display(el);	/* reset the display stuff */
-			ch_reset(el);	/* reset the input pointers */
-			read_clearmacros(&el->el_read->macros);
-			re_refresh(el); /* print the prompt again */
+			ch_reset(el, 1);	/* reset the input pointers */
+			re_refresh(el);	/* print the prompt again */
 			break;
 
 		case CC_ERROR:
 		default:	/* functions we don't know about */
-			terminal_beep(el);
-			terminal__flush(el);
+#ifdef DEBUG_READ
+			(void) fprintf(el->el_errfile,
+			    "*** editor ERROR ***\r\n\n");
+#endif /* DEBUG_READ */
+			term_beep(el);
+			term__flush();
 			break;
 		}
 		el->el_state.argument = 1;
@@ -601,21 +598,15 @@ el_wgets(EditLine *el, int *nread)
 			break;
 	}
 
-	terminal__flush(el);		/* flush any buffered output */
+	term__flush();		/* flush any buffered output */
 	/* make sure the tty is set up correctly */
 	if ((el->el_flags & UNBUFFERED) == 0) {
 		read_finish(el);
-		*nread = num != -1 ? num : 0;
-	} else
-		*nread = (int)(el->el_line.lastchar - el->el_line.buffer);
-
-	if (*nread == 0) {
-		if (num == -1) {
-			*nread = -1;
-			if (el->el_read->read_errno)
-				errno = el->el_read->read_errno;
-		}
-		return NULL;
-	} else
-		return el->el_line.buffer;
+		if (nread)
+			*nread = num;
+	} else {
+		if (nread)
+			*nread = el->el_line.lastchar - el->el_line.buffer;
+	}
+	return (num ? el->el_line.buffer : NULL);
 }
